import asyncio
import json
import logging
import os
from typing import Any, Union, List

import httpx
import requests

from creator.base.base_app import BaseApp
from creator.base.base_request import BaseRequest
from creator.base.base_response import BaseResponse, ResponseDataType, ResponseData
from creator.base.job_status import JobStatus
from creator.fcus_api.request_fcus_api import FcusTxt2Img, FcusUpscaleOrVary, FcusInpaintOrOutpaint
from creator.fcus_api.response_fcus_api import ResponseFcusApi

logger = logging.getLogger(__name__)


class AppFcusApi(BaseApp):
    param_classes = [FcusTxt2Img, FcusUpscaleOrVary, FcusInpaintOrOutpaint]
    output = {}

    def __init__(self):
        super().__init__()
        focus_port = os.environ.get("PORT_FCUS_API", 8888)
        if isinstance(focus_port, str):
            focus_port = int(focus_port)
        self.api_base_url = f"http://0.0.0.0:{focus_port}"

    async def create(self, params: Union[FcusTxt2Img, FcusUpscaleOrVary, FcusInpaintOrOutpaint]) -> BaseResponse:
        if isinstance(params, FcusTxt2Img):
            url = f"{self.api_base_url}/v1/generation/text-to-image"
        elif isinstance(params, FcusUpscaleOrVary):
            url = f"{self.api_base_url}/v1/generation/image-upscale-vary"
        elif isinstance(params, FcusInpaintOrOutpaint):
            url = f"{self.api_base_url}/v1/generation/image-inpaint-outpaint"
        else:
            raise ValueError("Invalid parameters for creation")

        data = json.dumps(params.__dict__)
        headers = {'Content-Type': 'application/json'}
        response = requests.post(url, data=data, headers=headers)
        # If we have a 422 here, print what the unprocessable entity is
        if response.status_code == 422:
            logger.warning(f"Unprocessable entity: {response.text}")
        response.raise_for_status()

        self.output = response.json()
        logger.debug(f"Response: {self.output}")
        job_id = self.output.get('job_id', None)
        output = None
        if job_id is not None:
            job_count = 0
            while output is None and job_count < 60:
                logger.debug(f"Job count: {job_count}")
                output = await self._get_output(job_id)
                job_count += 1
                if output is None:
                    await asyncio.sleep(1)
            response_data = ResponseData(data=output, data_type=ResponseDataType.IMAGE, total_count=len(output))
            return ResponseFcusApi(status="Job Finished", output=response_data, job_id=job_id)

    async def create_async(self, params: Union[FcusTxt2Img, FcusUpscaleOrVary, FcusInpaintOrOutpaint]) -> BaseResponse:
        if isinstance(params, FcusTxt2Img):
            model = params.base_model_name
            url = f"{self.api_base_url}/v1/generation/text-to-image"
        elif isinstance(params, FcusUpscaleOrVary):
            model = params.base_model_name
            url = f"{self.api_base_url}/v1/generation/image-upscale-vary"
        elif isinstance(params, FcusInpaintOrOutpaint):
            model = params.base_model_name
            url = f"{self.api_base_url}/v1/generation/image-inpaint-outpaint"
        else:
            raise ValueError("Invalid parameters for creation")
        if model is not None and model not in self.loaded_models:
            self.loaded_models.append(model)
        max_cached_models = os.environ.get("MAX_CACHED_MODELS", 3)
        if len(self.loaded_models) > max_cached_models:
            self.loaded_models = self.loaded_models[-max_cached_models:]

        data = json.dumps(params.model_dump())
        headers = {'Content-Type': 'application/json'}
        response = requests.post(url, data=data, headers=headers)
        # If we have a 422 here, print what the unprocessable entity is
        if response.status_code == 422:
            logger.warning(f"Unprocessable entity: {response.text}")
        response.raise_for_status()

        self.output = response.json()
        logger.debug(f"Response: {self.output}")
        job_id = self.output.get('job_id', None)
        if job_id is not None:
            return ResponseFcusApi.running(job_id)
        return ResponseFcusApi.error("Job ID not found", job_id)

    async def get_status(self, job_id: str) -> BaseResponse:
        output = await self._get_output(job_id)
        if output is None:
            return BaseResponse.error("Error getting job status")
        if not output:
            print(f"Job {job_id} is still running (or can't get output)")
            return BaseResponse.running(job_id)
        return BaseResponse.success(
            ResponseData(data=output, data_type=ResponseDataType.IMAGE, total_count=len(output)))

    async def upload_image(self, image_data: Any) -> str:
        pass

    async def get_models(self) -> List[str]:
        async with httpx.AsyncClient() as client:
            response = await client.get(f"{self.api_base_url}/v1/engines/all-models")
            response.raise_for_status()
            response_json = response.json()
            return response_json.get("model_filenames", [])

    async def test(self):
        async with httpx.AsyncClient() as client:
            url = f"{self.api_base_url}/docs"
            response = await client.get(url)
            if response.status_code == 200:
                return BaseResponse.active()
            else:
                return BaseResponse.error("Error testing connection")

    async def _get_output(self, job_id: str) -> Union[None, List[str]]:
        async with httpx.AsyncClient() as client:
            response = await client.get(f"{self.api_base_url}/v1/generation/query-job", params={'job_id': job_id})
            response.raise_for_status()
            response_json = response.json()
            status = response_json.get('job_status')
            logger.debug(f"Status: {status}")
            if status is None:
                return None
            if status != "Finished":
                return []
            outputs = [result.get('base64') for result in response_json.get('job_result', []) if result.get('finish_reason', "UNKNOWN") == "SUCCESS"]
            return outputs

    async def cache_model(self, model: str = "Juggernaut-XI-Prototype.safetensors") -> bool:
        # FCUS_API doesn't have a specific endpoint for caching models, so we just
        # Run a txt2img job with a very low resolution to load the model
        request = FcusTxt2Img().example()
        request.prompt = "foo"
        request.base_model_name = model
        request.aspect_ratios_selection = "64*64"
        response = await self.create(request)
        if response.status == JobStatus.FINISHED:
            return True
        return False


import asyncio
import json
import logging
import os
import traceback
from typing import Any, List, Union

import httpx

from creator.auto.request_auto import AutoTxt2Img, AutoImg2Img
from creator.auto.response_auto import ResponseAuto
from creator.base.base_app import BaseApp
from creator.base.base_response import BaseResponse, ResponseData, ResponseDataType
from creator.base.job_status import JobStatus

logger = logging.getLogger(__name__)


class AppAuto(BaseApp):
    param_classes = [AutoTxt2Img, AutoImg2Img]
    output = {}
    jobs = {}
    sched_jobs = {}

    def __init__(self):
        super().__init__()
        auto_port = os.environ.get("PORT_AUTO", 8081)
        if isinstance(auto_port, str):
            auto_port = int(auto_port)
        self.jobs = {}
        self.output = {}
        self.api_base_url = f"http://0.0.0.0:{auto_port}"
        app_config_path = "/opt/rd/apps/stable-diffusion-webui/config.json"
        if os.path.exists(app_config_path):
            with open(app_config_path, "r") as f:
                app_config = json.load(f)
                default_checkpoint = app_config.get("sd_model_checkpoint", None)
                if default_checkpoint:
                    self.loaded_models.append(default_checkpoint)

    async def create(self, params: Union[AutoTxt2Img, AutoImg2Img], job_id=None) -> BaseResponse:
        print(f"Creating job with params: {params}")
        endpoint_name = "txt2img" if isinstance(params, AutoTxt2Img) else "img2img"
        url = f"{self.api_base_url}/agent-scheduler/v1/{endpoint_name}"
        data = params.model_dump()
        checkpoint = data.get("sd_model_checkpoint", None)
        if checkpoint and checkpoint not in self.loaded_models:
            self.loaded_models.append(checkpoint)
            if len(self.loaded_models) > 3:
                self.loaded_models.pop(0)
        headers = {'Content-Type': 'application/json'}
        async with httpx.AsyncClient() as client:
            response = await client.post(url, data=data, headers=headers)
            response.raise_for_status()
        output = response.json()
        task_id = output.get("task_id", None)
        print(f"Response received: {output}")
        if task_id:
            self.jobs[task_id] = endpoint_name
            result = await self.get_status(task_id)
            while result.status == JobStatus.RUNNING:
                await asyncio.sleep(5)
            return result

    async def create_async(self, params: Union[AutoTxt2Img, AutoImg2Img]) -> BaseResponse:
        endpoint_name = "txt2img" if isinstance(params, AutoTxt2Img) else "img2img"
        url = f"{self.api_base_url}/agent-scheduler/v1/queue/{endpoint_name}"
        data = params.model_dump()
        headers = {'Content-Type': 'application/json'}
        checkpoint = data.get("sd_model_checkpoint", None)
        if checkpoint and checkpoint not in self.loaded_models:
            self.loaded_models.append(checkpoint)
            if len(self.loaded_models) > 3:
                self.loaded_models.pop(0)
        try:
            async with httpx.AsyncClient() as client:
                response = await client.post(url, data=data, headers=headers)
                response.raise_for_status()
            output = response.json()
            task_id = output.get("task_id", None)
            if task_id:
                return ResponseAuto.running(task_id)
        except Exception as e:
            logger.error(f"Error creating job: {e}")
            traceback.print_exc()
            return ResponseAuto.error(f"Error creating job: {e}")

    async def get_status(self, job_id: str) -> BaseResponse:
        try:
            url = f"{self.api_base_url}/agent-scheduler/v1/task/{job_id}/results"
            async with httpx.AsyncClient() as client:
                response = await client.get(url)
                response.raise_for_status()
            output = response.json()

            job_success = output.get("success", False)
            if not job_success:
                message = output.get("message", "Unknown status")
                if message == "Task is pending":
                    return ResponseAuto.running(job_id)
                else:
                    return ResponseAuto.error(f"Error getting job status: {message}", job_id)
            else:
                images = []
                data = output.get("data", [])
                if data and isinstance(data, list):
                    for output in data:
                        image = output.get("image", None)
                        if image:
                            # Remove the data:image/png;base64, prefix
                            image = image.split(",")[-1]
                            images.append(image)
                response_data = ResponseData(data=images, data_type=ResponseDataType.IMAGE, total_count=len(images))
                return ResponseAuto.success(response_data, job_id)

        except Exception as e:
            logger.error(f"Error getting job status for job ID {job_id}: {e}")
            traceback.print_exc()
            return ResponseAuto.error(f"Error getting job status: {e}", job_id)

    async def get_models(self) -> List[Any]:
        url = f"{self.api_base_url}/sdapi/v1/sd-models"
        async with httpx.AsyncClient() as client:
            response = await client.get(url)
            response.raise_for_status()
            response_json = response.json()
        models = [model.get("title") for model in response_json if model.get("title")]
        return models

    # Assuming you will provide more detail or use this method later
    async def upload_image(self, image_data: Any) -> str:
        pass

    async def test(self):
        auto_port = int(os.environ.get("PORT_AUTO", 8081))
        url = f"http://0.0.0.0:{auto_port}/docs"
        async with httpx.AsyncClient() as client:
            response = await client.get(url)
            if response.status_code == 200:
                return BaseResponse.active()
            else:
                return BaseResponse.error("Error testing connection")

    async def cache_model(self, model: str = "Juggernaut-XI-Prototype.safetensors") -> bool:
        request = {
            "sd_model_checkpoint": model
        }
        if model not in self.loaded_models:
            self.loaded_models.append(model)
        if len(self.loaded_models) > 3:
            self.loaded_models.pop(0)
        url = f"{self.api_base_url}/sdapi/v1/options"
        headers = {'Content-Type': 'application/json'}
        async with httpx.AsyncClient(timeout=300) as client:
            response = await client.post(url, json=request, headers=headers, timeout=300)
            response.raise_for_status()
            if response.status_code == 200:
                return True
            print(f"Error caching model: {response.text}")
        return False



import json

from pydantic import Field

from creator.base.base_request import BaseRequest

default_workflow_json = {
    "3": {
        "inputs": {
            "seed": 985289875893903,
            "steps": 20,
            "cfg": 7.5,
            "sampler_name": "dpmpp_2m_sde",
            "scheduler": "karras",
            "denoise": 1,
            "model": [
                "4",
                0
            ],
            "positive": [
                "6",
                0
            ],
            "negative": [
                "7",
                0
            ],
            "latent_image": [
                "5",
                0
            ]
        },
        "class_type": "KSampler",
        "_meta": {
            "title": "KSampler"
        }
    },
    "4": {
        "inputs": {
            "ckpt_name": "Juggernaut-XI-Prototype.safetensors",
            "key_opt": "",
            "mode": "Auto"
        },
        "class_type": "CheckpointLoaderSimpleShared //Inspire",
        "_meta": {
            "title": "Shared Checkpoint Loader (Inspire)"
        }
    },
    "5": {
        "inputs": {
            "width": 1024,
            "height": 768,
            "batch_size": 1
        },
        "class_type": "EmptyLatentImage",
        "_meta": {
            "title": "Empty Latent Image"
        }
    },
    "6": {
        "inputs": {
            "text": "a photograph of a smiling green alien holding a sign which says \"Run Diffusion\", ultra realistic, street, city, sci-fi, ultra realistic, cute, friendly",
            "clip": [
                "4",
                1
            ]
        },
        "class_type": "CLIPTextEncode",
        "_meta": {
            "title": "CLIP Text Encode (Prompt)"
        }
    },
    "7": {
        "inputs": {
            "text": "blurry, low quality",
            "clip": [
                "4",
                1
            ]
        },
        "class_type": "CLIPTextEncode",
        "_meta": {
            "title": "CLIP Text Encode (Prompt)"
        }
    },
    "8": {
        "inputs": {
            "samples": [
                "3",
                0
            ],
            "vae": [
                "4",
                2
            ]
        },
        "class_type": "VAEDecode",
        "_meta": {
            "title": "VAE Decode"
        }
    },
    "9": {
        "inputs": {
            "filename_prefix": "ComfyUI",
            "images": [
                "8",
                0
            ]
        },
        "class_type": "SaveImage",
        "_meta": {
            "title": "Save Image"
        }
    }
}

default_workflow_json_2 = {
    "21": {
        "inputs": {
            "width": 1920,
            "height": 1080,
            "batch_size": 2
        },
        "class_type": "EmptyLatentImage",
        "_meta": {
            "title": "Empty Latent Image"
        }
    },
    "67": {
        "inputs": {
            "samples": [
                "132",
                0
            ],
            "vae": [
                "135",
                2
            ]
        },
        "class_type": "VAEDecode",
        "_meta": {
            "title": "VAE Decode"
        }
    },
    "99": {
        "inputs": {
            "seed": 1011799137402762
        },
        "class_type": "CR Seed",
        "_meta": {
            "title": "🌱 CR Seed"
        }
    },
    "100": {
        "inputs": {
            "text": [
                "114:0",
                0
            ],
            "seed": [
                "99",
                0
            ],
            "log_prompt": "No"
        },
        "class_type": "PromptExpansion",
        "_meta": {
            "title": "Prompt Expansion"
        }
    },
    "101": {
        "inputs": {
            "text": [
                "144",
                0
            ]
        },
        "class_type": "ShowText|pysssss",
        "_meta": {
            "title": "Show Text 🐍"
        }
    },
    "102": {
        "inputs": {
            "text": [
                "145",
                0
            ]
        },
        "class_type": "ShowText|pysssss",
        "_meta": {
            "title": "Show Text 🐍"
        }
    },
    "120": {
        "inputs": {
            "b1": 1.01,
            "b2": 1.02,
            "s1": 0.99,
            "s2": 0.9500000000000001,
            "model": [
                "135",
                0
            ]
        },
        "class_type": "FreeU_V2",
        "_meta": {
            "title": "FreeU_V2"
        }
    },
    "132": {
        "inputs": {
            "add_noise": False,
            "noise_seed": [
                "99",
                0
            ],
            "steps": 15,
            "cfg": 4,
            "sampler_name": "euler_ancestral",
            "scheduler": "karras",
            "start_at_step": 0,
            "end_at_step": 10000,
            "noise_mode": "CPU",
            "return_with_leftover_noise": False,
            "batch_seed_mode": "incremental",
            "variation_seed": 0,
            "variation_strength": 0,
            "model": [
                "134",
                0
            ],
            "positive": [
                "138:1",
                0
            ],
            "negative": [
                "138:0",
                0
            ],
            "latent_image": [
                "21",
                0
            ]
        },
        "class_type": "KSamplerAdvanced //Inspire",
        "_meta": {
            "title": "KSamplerAdvanced (inspire)"
        }
    },
    "134": {
        "inputs": {
            "ds_depth_1": 3,
            "ds_depth_2": 3,
            "ds_timestep_1": 900,
            "ds_timestep_2": 650,
            "resize_scale_1": 2,
            "resize_scale_2": 2,
            "model": [
                "120",
                0
            ]
        },
        "class_type": "Hires",
        "_meta": {
            "title": "Apply Kohya's HiresFix"
        }
    },
    "135": {
        "inputs": {
            "ckpt_name": "Juggernaut-XI-Prototype.safetensors",
            "key_opt": "",
            "mode": "Auto"
        },
        "class_type": "CheckpointLoaderSimpleShared //Inspire",
        "_meta": {
            "title": "Shared Checkpoint Loader (Inspire)"
        }
    },
    "139": {
        "inputs": {
            "prompt": "three squirrels in business suits having a meeting at a tiny table in a park"
        },
        "class_type": "CR Prompt Text",
        "_meta": {
            "title": "⚙️ CR Prompt Text"
        }
    },
    "140": {
        "inputs": {
            "prompt": "blurry, bad eyes, bad hands, deformed, low quality"
        },
        "class_type": "CR Prompt Text",
        "_meta": {
            "title": "⚙️ CR Prompt Text"
        }
    },
    "144": {
        "inputs": {
            "action": "replace",
            "tidy_tags": "yes",
            "text_a": [
                "100",
                0
            ],
            "text_b": "",
            "text_c": "",
            "result": "three squirrels in business suits having a meeting at a tiny table in a park, professional photograph, bokeh, 8k detailed, 35mm, vintage, highly detailed, found footage, cinematic still, emotional, harmonious, vignette, 4k detailed, sharp focus, high budget, epic, gorgeous, film grain, three squirrels in business suits having a meeting at a tiny table in a park, professional photograph, bokeh, 8k detailed, 35mm, vintage, highly detailed, found footage, cinematic still, emotional, harmonious, vignette, 4k detailed, sharp focus, high budget, epic, gorgeous, film grain, atmosphere, bright, rich vivid colors"
        },
        "class_type": "StringFunction|pysssss",
        "_meta": {
            "title": "String Function 🐍"
        }
    },
    "145": {
        "inputs": {
            "action": "replace",
            "tidy_tags": "yes",
            "text_a": [
                "114:1",
                0
            ],
            "text_b": "",
            "text_c": "",
            "result": "blurry, bad eyes, bad hands, deformed, low quality, blurry, cropped, bad face, saturated, contrast, deformed iris, deformed pupils, semi-realistic, cgi, 3d, render, sketch, cartoon, drawing, anime, text, cropped, out of frame, worst quality, low quality, jpeg artifacts, ugly, duplicate, morbid, mutilated, extra fingers, mutated hands, poorly drawn hands, poorly drawn face, mutation, deformed, dehydrated, bad anatomy, bad proportions, extra limbs, cloned face, disfigured, gross proportions, malformed limbs, missing arms, missing legs, extra arms, extra legs, fused fingers, too many fingers, long neck, anime, cartoon, graphic, (blur, blurry, bokeh), text, painting, crayon, graphite, abstract, glitch, deformed, mutated, ugly, disfigured, deformed, bad anatomy, disfigured, poorly drawn face, mutated, extra limb, ugly, poorly drawn hands, missing limb, floating limbs, disconnected limbs, disconnected head, malformed hands, long neck, mutated hands and fingers, bad hands, missing fingers, cropped, worst quality, low quality, mutation, poorly drawn, huge calf, bad hands, fused hand, missing hand, disappearing arms, disappearing thigh, disappearing calf, disappearing legs, missing fingers, fused fingers, abnormal eye proportion, Abnormal hands, abnormal legs, abnormal feet, abnormal fingers, drawing, painting, crayon, sketch, graphite, impressionist, noisy, blurry, soft, deformed, ugly, anime, cartoon, graphic, text, painting, crayon, graphite, abstract, glitch"
        },
        "class_type": "StringFunction|pysssss",
        "_meta": {
            "title": "String Function 🐍"
        }
    },
    "148": {
        "inputs": {
            "filename_prefix": [
                "139",
                0
            ],
            "images": [
                "67",
                0
            ]
        },
        "class_type": "SaveImage",
        "_meta": {
            "title": "Save Image"
        }
    },
    "119:0": {
        "inputs": {
            "text_positive": "",
            "text_negative": "",
            "style": "Fooocus Photograph",
            "log_prompt": True,
            "style_positive": True,
            "style_negative": True
        },
        "class_type": "SDXLPromptStyler",
        "_meta": {
            "title": "SDXL Prompt Styler"
        }
    },
    "119:1": {
        "inputs": {
            "text_positive": "",
            "text_negative": "",
            "style": "Fooocus Sharp",
            "log_prompt": True,
            "style_positive": True,
            "style_negative": True
        },
        "class_type": "SDXLPromptStyler",
        "_meta": {
            "title": "SDXL Prompt Styler"
        }
    },
    "119:2": {
        "inputs": {
            "text_positive": "",
            "text_negative": "",
            "style": "Fooocus Negative",
            "log_prompt": True,
            "style_positive": True,
            "style_negative": True
        },
        "class_type": "SDXLPromptStyler",
        "_meta": {
            "title": "SDXL Prompt Styler"
        }
    },
    "133:1": {
        "inputs": {
            "text1": [
                "140",
                0
            ],
            "text2": [
                "119:0",
                1
            ],
            "separator": ", "
        },
        "class_type": "CR Text Concatenate",
        "_meta": {
            "title": "🔤 CR Text Concatenate"
        }
    },
    "133:2": {
        "inputs": {
            "text1": [
                "119:1",
                0
            ],
            "text2": [
                "119:2",
                0
            ],
            "separator": ", "
        },
        "class_type": "CR Text Concatenate",
        "_meta": {
            "title": "🔤 CR Text Concatenate"
        }
    },
    "133:3": {
        "inputs": {
            "text1": [
                "119:1",
                1
            ],
            "text2": [
                "119:2",
                1
            ],
            "separator": ", "
        },
        "class_type": "CR Text Concatenate",
        "_meta": {
            "title": "🔤 CR Text Concatenate"
        }
    },
    "133:0": {
        "inputs": {
            "text1": [
                "139",
                0
            ],
            "text2": [
                "119:0",
                0
            ],
            "separator": ", "
        },
        "class_type": "CR Text Concatenate",
        "_meta": {
            "title": "🔤 CR Text Concatenate"
        }
    },
    "114:1": {
        "inputs": {
            "text1": [
                "133:1",
                0
            ],
            "text2": [
                "133:3",
                0
            ],
            "separator": ", "
        },
        "class_type": "CR Text Concatenate",
        "_meta": {
            "title": "🔤 CR Text Concatenate"
        }
    },
    "114:0": {
        "inputs": {
            "text1": [
                "133:0",
                0
            ],
            "text2": [
                "133:2",
                0
            ],
            "separator": ", "
        },
        "class_type": "CR Text Concatenate",
        "_meta": {
            "title": "🔤 CR Text Concatenate"
        }
    },
    "138:0": {
        "inputs": {
            "text": [
                "102",
                0
            ],
            "clip": [
                "135",
                1
            ]
        },
        "class_type": "CLIPTextEncode",
        "_meta": {
            "title": "CLIP Text Encode (Prompt)"
        }
    },
    "138:1": {
        "inputs": {
            "text": [
                "101",
                0
            ],
            "clip": [
                "135",
                1
            ]
        },
        "class_type": "CLIPTextEncode",
        "_meta": {
            "title": "CLIP Text Encode (Prompt)"
        }
    }
}

model_load_json = {
    "1": {
        "inputs": {
            "ckpt_name": "Juggernaut-XI-Prototype.safetensors",
            "key_opt": "",
            "mode": "Auto"
        },
        "class_type": "CheckpointLoaderSimpleShared //Inspire",
        "_meta": {
            "title": "Shared Checkpoint Loader (Inspire)"
        }
    },
    "3": {
        "inputs": {
            "text": "",
            "clip": [
                "1",
                1
            ]
        },
        "class_type": "CLIPTextEncode",
        "_meta": {
            "title": "CLIP Text Encode (Prompt)"
        }
    },
    "4": {
        "inputs": {
            "root_dir": "output",
            "file": "file.txt",
            "append": "append",
            "insert": True,
            "text": [
                "10",
                2
            ]
        },
        "class_type": "SaveText|pysssss",
        "_meta": {
            "title": "Save Text"
        }
    },
    "10": {
        "inputs": {
            "latent_suffix": "6934042_cache",
            "image_suffix": "8568828_cache",
            "conditioning_suffix": "91602225_cache",
            "output_path": "/opt/rd/apps/ComfyUI/custom_nodes/was-node-suite-comfyui/cache",
            "conditioning": [
                "3",
                0
            ]
        },
        "class_type": "Cache Node",
        "_meta": {
            "title": "Cache Node"
        }
    }
}


class CmfyWorkflow(BaseRequest):
    workflow_json: str = Field("", examples=[json.dumps(default_workflow_json)])


class CmfyWorkflowFcus(BaseRequest):
    workflow_json: str = Field("", examples=[json.dumps(default_workflow_json_2)])



import base64
from enum import Enum
from typing import List, Optional, Any
from fastapi.responses import Response

from pydantic import BaseModel, Field

from creator.base.job_status import JobStatus


class ResponseDataType(str, Enum):
    IMAGE = "image"
    VIDEO = "video"
    GIF = "gif"
    TEXT = "text"
    AUDIO = "audio"
    NONE = "none"


class ResponseData(BaseModel):
    data: List[Any] = Field(default_factory=list, description="The data to be returned.")
    data_type: ResponseDataType = Field(ResponseDataType.NONE, description="The type of data being returned.")
    total_count: int = 0


class BaseResponse(BaseModel):
    status: JobStatus = JobStatus.PENDING
    output: Optional[ResponseData] = None
    error_msg: Optional[str] = None
    job_id: Optional[str] = None

    # Override Pydantic's dict method to exclude methods
    def dict(self, **kwargs):
        return super().dict(
            **kwargs,
            exclude={
                "parse_response",
                "success",
                "error"
            }
        )

    @classmethod
    def parse_response(cls, response: dict, job_id: str = None):
        """Parse the response from the API."""
        return cls(**response, job_id=job_id)

    @classmethod
    def success(cls, data: ResponseData, job_id: str = None):
        """Return a successful response."""
        return cls(status=JobStatus.FINISHED, output=data, job_id=job_id)

    @classmethod
    def active(cls, job_id: str = None):
        return cls(status=JobStatus.READY, job_id=job_id)

    @classmethod
    def running(cls, job_id: str = None):
        """Return a pending response."""
        return cls(status=JobStatus.RUNNING, job_id=job_id)

    @classmethod
    def error(cls, error: str, job_id: str = None):
        """Return an error response."""
        return cls(status=JobStatus.FAILED, error_msg=error, job_id=job_id)

    def image_response(self, job_id: str = None):
        """Return a response containing images if output is image data."""
        if self.output and self.output.data_type == ResponseDataType.IMAGE:
            image_data_list = self.output.data
            if len(image_data_list) == 1:
                # If only one image, return a single image response
                image_data = base64.b64decode(image_data_list[0])
                return Response(content=image_data, media_type="image/png")
            else:
                # If multiple images, construct a multipart response
                boundary = "boundary_separator"
                content_type = f"multipart/mixed; boundary={boundary}"
                body_parts = []
                for image_data in image_data_list:
                    image_bytes = base64.b64decode(image_data)
                    image_part = (
                        f"\r\n--{boundary}\r\n"
                        f"Content-Type: image/png\r\n"
                        f"Content-Length: {len(image_bytes)}\r\n\r\n"
                    )
                    body_parts.append(image_part.encode("utf-8"))
                    body_parts.append(image_bytes)
                body = b"".join(body_parts) + f"\r\n--{boundary}--\r\n".encode("utf-8")
                return Response(content=body, media_type=content_type)
        else:
            return self



from typing import Dict, Any
from creator.base.base_response import BaseResponse, ResponseData, ResponseDataType


class ResponseAuto(BaseResponse):
    @classmethod
    def parse_response(cls, response: Dict[str, Any], job_id: str = None):
        data = response.get('images', [])
        if not data:
            print(f"No images found in response: {response}")
            return cls.error("No images found in response")
        print("Images found in response")
        response_data = ResponseData(data=data, data_type=ResponseDataType.IMAGE, total_count=len(data))
        return cls.success(response_data, job_id=job_id)
